from six.moves import reduce

import operator


# reversible_operators = [
#     ["__add__", "__radd__"],
#     ["__sub__", "__rsub__"],
#     ["__mul__", "__rmul__"],
#     ["__floordiv__", "__rfloordiv__"],
#     ["__div__", "__rdiv__"],
#     ["__truediv__", "__rtruediv__"],
#     ["__mod__", "__rmod__"],
#     ["__divmod__", "__rdivmod__"],
#     ["__pow__", "__rpow__"],
#     ["__lshift__", "__rlshift__"],
#     ["__rshift__", "__rrshift__"],
#     ["__and__", "__rand__"],
#     ["__or__", "__ror__"],
#     ["__xor__", "__rxor__"],
# ]

# normal_operators = [
#     "__abs__", "__concat__", "__contains__", "__delitem__", "__delslice__",
#     "__eq__", "__file__", "__ge__", "__getitem__", "__getslice__", "__gt__", 
#     "__iadd__", "__iand__", "__iconcat__", "__idiv__", "__ifloordiv__", 
#     "__ilshift__", "__imod__", "__imul__", "__index__", "__inv__", "__invert__",
#     "__ior__", "__ipow__", "__irepeat__", "__irshift__", "__isub__", 
#     "__itruediv__", "__ixor__", "__le__", "__lt__", "__ne__", "__neg__",
#     "__not__", "__package__", "__pos__", "__repeat__", "__setitem__",
#     "__setslice__", "__radd__", "__rsub__", "__rmul__", "__rfloordiv__",
#     "__rdiv__", "__rtruediv__", "__rmod__", "__rdivmod__", "__rpow__", 
#     "__rlshift__",  "__rand__",  "__ror__",  "__rxor__",  # "__rrshift__",
# ]

class Operator(object):

  def __init__(self, name, symbol, precedence):
    self.name = name
    self.format_string = self.get_format_string(symbol)
    self.precedence = precedence

  def apply(self, *args, **kwargs):
    func = getattr(operator, self.name)
    return func(*args)

  def format_exp(self, exp):
    if isinstance(exp, Later) and exp._step.precedence < self.precedence:
      return "({0})".format(str(exp))
    else:
      return str(exp)

  def format_operation(self, *args, **kwargs):
    return self.format_string.format(*args)


class BinaryOperator(Operator):

  def apply(self, *args, **kwargs):
    func = getattr(operator, self.name)
    return func(*args)

  def get_format_string(self, symbol):
    return "{0} " + symbol + " {1}"


class ReverseBinaryOperator(Operator):
  
  def apply(self, *args, **kwargs):
    # Note: this is a binary operator, so we know there are only two args.
    # This is the reverse, so we have to flip them.
    func = getattr(operator, self.name)
    return func(args[1], args[0])

  def get_format_string(self, symbol):
    return "{1} " + symbol + " {0}"


class UnaryOperator(Operator):

  def get_format_string(self, symbol):
    return symbol + "{0}"


class Step(object):

  def __init__(self, precedence, *args, **kwargs):
    self.precedence = precedence
    self.args = args
    self.kwargs = kwargs

  def format_operation(self, obj):
    raise NotImplementedError

  def format_exp(self, exp):
    if isinstance(exp, Later) and exp._step.precedence < self.precedence:
      return "({0})".format(str(exp))
    else:
      return str(exp)

  def evaluated_args(self, df, **kwargs):
    return [(arg.evaluate(df, **kwargs) if isinstance(arg, Later) else arg)
            for arg in self.args]

  def evaluated_kwargs(self, df, **kwargs):
    return {k: v.evaluate(df, **kwargs) if isinstance(v, Later) else v
            for k, v in self.kwargs.items()}


class OperatorStep(Step):

  def __init__(self, operator, *args, **kwargs):
    self.operator = operator
    super(OperatorStep, self).__init__(operator.precedence, 
                                       *args, 
                                       **kwargs)

  def evaluate(self, previousResult, original, **kwargs):
    return self.operator.apply(previousResult, 
                               *self.evaluated_args(original, **kwargs),
                               **self.evaluated_kwargs(original, **kwargs))

  def format_operation(self, obj):
    formatted_args = [self.format_exp(arg) for arg in [obj] + list(self.args)]
    return self.operator.format_operation(*formatted_args, **self.kwargs)


class ArgStep(Step):

  def format_args(self):
      # We sort here because keyword arguments get arbitrary ordering inside the 
      # function call. Support PEP 0468 to help fix this issue!
      # https://www.python.org/dev/peps/pep-0468/
      kwarg_strs = sorted(["{0}={1}".format(k, _addQuotes(v)) 
                            for k, v in self.kwargs.items()])
      arg_strs = list(map(str, self.args))
      full_str = ", ".join(arg_strs + kwarg_strs)
      return full_str


class CallStep(ArgStep):

  def __init__(self, *args, **kwargs):
    super(CallStep, self).__init__(15, *args, **kwargs)

  def format_operation(self, obj):
    return "{0}({1})".format(obj,
                             self.format_args())

  def evaluate(self, previousResult, original, **kwargs):
    return previousResult.__call__(*self.evaluated_args(original, **kwargs),
                                   **self.evaluated_kwargs(original, **kwargs))


class FunctionStep(ArgStep):

  def __init__(self, func, *args, **kwargs):
    self.func = func
    super(FunctionStep, self).__init__(15, *args, **kwargs)

  def format_operation(self, obj):
    return "{0}({1})".format(self.func.__name__,
                             self.format_args())

  def evaluate(self, previousResult, original, **kwargs):
    return self.func(*self.evaluated_args(original, **kwargs),
                     **self.evaluated_kwargs(original, **kwargs))


class AttributeStep(Step):

  def __init__(self, attr):
    self.attr = attr
    super(AttributeStep, self).__init__(15)

  def format_operation(self, obj):
    return "{0}.{1}".format(self.format_exp(obj),
                            self.attr)

  def evaluate(self, previousResult, original, **kwargs):
    return getattr(previousResult, self.attr)


class BracketStep(Step):

  def __init__(self, key):
    self.key = key
    super(BracketStep, self).__init__(15)

  def format_operation(self, obj):
    return '{0}[{1}]'.format(self.format_exp(obj),
                             self.format_arg())

  def format_arg(self):
    if isinstance(self.key, slice):
      _str = ""
      (start, stop, step) = (self.key.start, self.key.stop, self.key.step)
      if start is not None:
        _str += str(start)
      _str += ":"
      if stop is not None:
        _str += str(stop)
      if step is not None:
        _str += ":" + str(step)
      return _str
    elif isinstance(self.key, str):
      return '"{0}"'.format(self.key)
    else:
      return str(self.key)

  def evaluate(self, previousResult, original, **kwargs):
    return operator.getitem(previousResult, self.key)


class SliceStep(BracketStep):
  """Necessary for compatibility."""

  def format_operation(self, obj, args, kwargs):
    return "{0}[{1}]".format(self.format_exp(obj),
                             self.format_arg(slice(args[0], args[1])))


class IdentityStep(Step):

  def __init__(self):
    super(IdentityStep, self).__init__(17)

  def evaluate(self, previousResult, original, **kwargs):
    return previousResult

  def format_operation(self, obj):
    return "X._"


OPERATORS = {
  # see https://docs.python.org/2/reference/expressions.html#operator-precedence
  "__lt__": BinaryOperator("__lt__", "<", 6),
  "__le__": BinaryOperator("__le__", "<=", 6),
  "__eq__": BinaryOperator("__eq__", "==", 6),
  "__ne__": BinaryOperator("__ne__", "!=", 6),
  "__ge__": BinaryOperator("__ge__", ">=", 6),
  "__gt__": BinaryOperator("__gt__", ">", 6),
  "__or__": BinaryOperator("__or__", "|", 7),
  "__ror__": ReverseBinaryOperator("__or__", "|", 7),
  "__xor__": BinaryOperator("__xor__", "^", 8),
  "__rxor__": ReverseBinaryOperator("__xor__", "^", 8),
  "__and__": BinaryOperator("__and__", "&", 9),
  "__rand__": ReverseBinaryOperator("__and__", "&", 9),
  "__lshift__": BinaryOperator("__lshift__", "<<", 10),
  "__rlshift__": ReverseBinaryOperator("__lshift__", "<<", 10),
  "__rshift__": BinaryOperator("__rshift__", ">>", 10),
  "__rrshift__": ReverseBinaryOperator("__rshift__", ">>", 10),
  "__add__": BinaryOperator("__add__", "+", 11),
  "__radd__": ReverseBinaryOperator("__add__", "+", 11),
  "__sub__": BinaryOperator("__sub__", "-", 11),
  "__rsub__": ReverseBinaryOperator("__sub__", "-", 11),
  "__mul__": BinaryOperator("__mul__", "*", 12),
  "__rmul__": ReverseBinaryOperator("__mul__", "*", 12),
  "__div__": BinaryOperator("__div__", "/", 12),
  "__rdiv__": ReverseBinaryOperator("__div__", "/", 12),
  "__truediv__": BinaryOperator("__truediv__", "/", 12),
  "__rtruediv__": ReverseBinaryOperator("__truediv__", "/", 12),
  "__floordiv__": BinaryOperator("__floordiv__", "//", 12),
  "__rfloordiv__": ReverseBinaryOperator("__floordiv__", "//", 12),
  "__mod__": BinaryOperator("__mod__", "%", 12),
  "__rmod__": ReverseBinaryOperator("__mod__", "%", 12),
  "__neg__": UnaryOperator("__neg__", "-", 13),
  "__pos__": UnaryOperator("__pos__", "+", 13),
  "__invert__": UnaryOperator("__invert__", "~", 13),
  "__pow__": BinaryOperator("__pow__", "**", 14),
  "__rpow__": ReverseBinaryOperator("__pow__", "**", 14),
}

def add_operator_hooks(cls):

  def get_hook(name):
    operator = OPERATORS[name]
    def op_hook(self, *args, **kwargs):
      return Later(OperatorStep(operator, *args, **kwargs), self)
    return op_hook

  for name in OPERATORS:
    setattr(cls, name, get_hook(name))

  return cls


def _addQuotes(item):
  return '"' + item + '"' if isinstance(item, str) else item


@add_operator_hooks
class Later(object):
  """Object which represents a computation to be carried out later.

  The Later object allows us to save computation that cannot currently be 
  executed. It will later receive a DataFrame as an input, and all computation 
  will be carried out upon this DataFrame object.

  Thus, we can refer to columns of the DataFrame as inputs to functions without 
  having the DataFrame currently available:
  In : diamonds >> sift(X.carat > 4) >> select(X.carat, X.price)
  Out:
         carat  price
  25998   4.01  15223
  25999   4.01  15223
  27130   4.13  17329
  27415   5.01  18018
  27630   4.50  18531

  The special Later name, "_" will refer to the entire DataFrame. For example, 
  In: diamonds >> sample_n(6) >> select(X.carat, X.price) >> X._.T
  Out:
           18966    19729   9445   49951    3087    33128
  carat     1.16     1.52     0.9    0.3     0.74    0.31
  price  7803.00  8299.00  4593.0  540.0  3315.00  816.00
  """

  def __init__(self, step, previous, name = None):
    if isinstance(step, IdentityStep) or name is not None:
      self._queue = []
    else:
      self._queue = previous._queue
    self._step = step
    self._queue.append(step)
    self._previous = previous
    self._name = name

  def evaluate(self, previousResult, original=None, special=None):
    if special not in {None, "transform", "agg"}:
      raise Exception("Special must be one of None, 'transform', or 'agg'.")

    original = original if original is not None else previousResult

    if (original._grouped_self
        and special 
        and not isinstance(self._queue[0], FunctionStep)):
      name = GetName(self)

      # TODO: Rewrite this, this is terrible.
      go_to_index = len(self._queue)
      for idx, item in enumerate(self._queue):
        if isinstance(item, OperatorStep):
          go_to_index = idx
          break

      transform_input = lambda x: reduce(
          lambda prevResult, f: f.evaluate(prevResult, original, special=special),
          self._queue[1:go_to_index],
          x
      )
      if special == "transform":
        out = original._grouped_self[name].transform(transform_input)
      elif special == "agg":
        out = original._grouped_self[name].agg(transform_input)
      out = reduce(lambda prevResult, f: f.evaluate(prevResult, original, special=special),
                      self._queue[go_to_index:],
                      out)
      return out
    else:
      output = reduce(lambda prevResult, f: f.evaluate(prevResult, original),
                      self._queue,
                      original)
      return output
    
  def __str__(self):
    return self._step.format_operation(self._previous)

  def __repr__(self):
    return str(self)

  def __getattr__(self, attr):
    return Later(AttributeStep(attr), self)

  def __getitem__(self, attr):
    return Later(BracketStep(attr), self)

  def __call__(self, *args, **kwargs):
    return Later(CallStep(*args, **kwargs), self)

  def __rrshift__(self, df):
    otherDf = DplyFrame(df.copy(deep=True))
    return self.evaluate(otherDf)


def CreateLaterFunction(fcn, *args, **kwargs):
  return Later(FunctionStep(fcn, *args, **kwargs), IdentityStep(), "function")

def DelayFunction(fcn):
  def DelayedFcnCall(*args, **kwargs):
    # Check to see if any args or kw are Later. If not, return normal fcn.
    if (len([a for a in args if isinstance(a, Later)]) == 0 and
        len([v for k, v in kwargs.items() if isinstance(v, Later)]) == 0):
      return fcn(*args, **kwargs)
    else:
      return CreateLaterFunction(fcn, *args, **kwargs)

  return DelayedFcnCall


def GetName(later):
  while not later._name:
    later = later._previous
  return later._name


class Manager(object):
  """Object which helps create a delayed computational unit.

  Typically will be set as a global variable X.
  X.foo will refer to the "foo" column of the DataFrame in which it is later
  applied. 

  Manager can be used in two ways: 
  (1) attribute notation: X.foo
  (2) item notation: X["foo"]

  Attribute notation is preferred but item notation can be used in cases where 
  column names contain characters on which python will choke, such as spaces, 
  periods, and so forth.
  """
  def __getattr__(self, attr):
    if attr == "_":
      return Later(IdentityStep(), self, "_")
    else:
      return Later(AttributeStep(attr), self, attr)

  def __getitem__(self, key):
    if key == "_":
      return Later(IdentityStep(), self, "_")
    else:
      return Later(BracketStep(key), self, key)

  def evaluate(self, previousResult, original, **kwargs):
    return original

  def __str__(self):
    return "X"

  def __repr__(self):
    return str(self)


X = Manager()






from timeit import Timer

setup = """
from dplython import X, mutate, group_by, diamonds
diamonds = diamonds >> mutate(bin=X["Unnamed: 0"] % 5000)
gbinp = diamonds.groupby("bin")
gbind = diamonds >> group_by(X.bin)
"""

setup2 = """
import pandas as pd
from dplython import X, mutate, group_by, diamonds
gbinp = pd.DataFrame(diamonds).groupby("carat")
"""
# diamonds = diamonds >> mutate(bin=X["Unnamed: 0"] % 5000)
# gbind = diamonds >> group_by(X.bin)

test_stmt = "7 + 7"

Timer(stmt=test_stmt,
      setup=setup).timeit()

Timer(stmt="gbinp.x.transform('mean') + gbinp.y.transform('mean')",
      setup=setup).timeit()
Timer(stmt="gbind >> mutate(foo=X.x.mean() + X.y.mean())",
      setup=setup).timeit()


# Chris Riederer
# 2016-02-21

"""Testing for python dplyr."""

from collections import Counter
import math
import unittest
import os

# import this for quick dataframe creation
import sys
if sys.version_info[0] < 3:
  from StringIO import StringIO
else:
  from io import StringIO

import numpy as np
import numpy.testing as npt
import pandas as pd

from dplython import *


def load_diamonds():
    root = os.path.abspath(os.path.dirname(__file__))
    path = os.path.join(root, 'data', 'diamonds.csv')
    return DplyFrame(pd.read_csv(path))


class TestLaterStrMethod(unittest.TestCase):
  
  def test_column_name(self):
    foo = X.foo
    self.assertEqual(foo._str, 'X.foo.')

  def test_str(self):
    foo = X.foo
    self.assertEqual(str(foo), 'X.foo')

  def test_bracket(self):
    foo = X["foo"]
    self.assertEqual(str(foo), 'X["foo"]')

  def test_later_with_method(self):
    foo = X.foo.mean()
    self.assertEqual(str(foo), 'X.foo.mean()')

  def test_later_with_method_call(self):
    foo = X.foo.mean()
    self.assertEqual(str(foo), 'X.foo.mean()')
    foo = X.foo.mean(1)
    self.assertEqual(str(foo), 'X.foo.mean(1)')
    foo = X.foo.mean(1, 2)
    self.assertEqual(str(foo), 'X.foo.mean(1, 2)')
    foo = X.foo.mean(numeric_only=True)
    self.assertEqual(str(foo), 'X.foo.mean(numeric_only=True)')
    # The order is different here, because the original order of the kwargs is
    # lost when kwargs are passed to the function. To insure consistent results,
    #  the kwargs are sorted alphabetically by key. To help deal with this
    # issue, support PEP 0468: https://www.python.org/dev/peps/pep-0468/
    foo = X.foo.mean(numeric_only=True, level="bar")
    self.assertEqual(str(foo), 'X.foo.mean(level="bar", '
                               'numeric_only=True)')
    foo = X.foo.mean(1, numeric_only=True, level="bar")
    self.assertEqual(str(foo), 'X.foo.mean(1, level="bar", '
                               'numeric_only=True)')
    foo = X.foo.mean(1, 2, numeric_only=True, level="bar")
    self.assertEqual(str(foo), 'X.foo.mean(1, 2, level="bar", '
                               'numeric_only=True)')
    foo = X.foo.mean(X.y.mean())
    self.assertEqual(str(foo), 'X.foo.mean('
                               'X.y.mean())')

  def test_later_with_delayed_function(self):
    mylen = DelayFunction(len)
    foo = mylen(X.foo)
    self.assertEqual(str(foo), 'len(X.foo)')

  def test_more_later_ops_str(self):
    mylen = DelayFunction(len)
    foo = -mylen(X.foo) + X.y.mean() // X.y.median()
    self.assertEqual(str(foo), '-len(X.foo) + '
                               'X.y.mean() // '
                               'X.y.median()')
    bar = -(mylen(X.bar) + X.y.mean()) * X.y.median()
    self.assertEqual(str(bar), '-(len(X.bar) + X.y.mean()) * '
                               'X.y.median()')
    baz = 6 + (X.y.mean() % 4) - X.bar.sum()
    self.assertEqual(str(baz), '6 + X.y.mean() % 4 - X.bar.sum()')
    buzz = (X.bar / 4) == X.baz
    self.assertEqual(str(buzz), 'X.bar / 4 == X.baz')
    biz = X.foo[4] / X.bar[2:3] + X.baz[::2]
    self.assertEqual(str(biz), 'X.foo[4] / X.bar[2:3] + X.baz[::2]')


class TestDelayFunctions(unittest.TestCase):
  diamonds = load_diamonds()

  def test_function_args(self):
    foo_pd = PairwiseGreater(self.diamonds["x"], self.diamonds["y"])
    foo_dp = self.diamonds >> mutate(foo=PairwiseGreater(X.x, X.y)) >> X._.foo
    self.assertTrue((foo_pd == foo_dp).all())

  def test_function_kwargs(self):
    @DelayFunction
    def PairwiseGreaterKwargs(series1=None, series2=None):
      index = series1.index
      newSeries = pd.Series(np.zeros(len(series1)))
      s1_ind = series1 > series2
      s2_ind = series1 <= series2
      newSeries[s1_ind] = series1[s1_ind]
      newSeries[s2_ind] = series2[s2_ind]
      newSeries.index = index
      return newSeries

    foo_pd = PairwiseGreaterKwargs(
        series1=self.diamonds["x"], series2=self.diamonds["y"])
    foo_dp = (self.diamonds >>
                mutate(foo=PairwiseGreaterKwargs(series1=X.x, series2=X.y)) >>
                X._.foo)
    self.assertTrue((foo_pd == foo_dp).all())


class TestMutates(unittest.TestCase):
  diamonds = load_diamonds()

  def test_equality(self):
    self.assertTrue(self.diamonds.equals(self.diamonds))

  def test_addconst(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd["ones"] = 1
    diamonds_dp = self.diamonds >> mutate(ones=1)
    self.assertTrue(diamonds_pd.equals(diamonds_dp))

  def test_newcolumn(self):
    diamonds_pd = self.diamonds.copy()
    newcol = range(len(diamonds_pd))
    diamonds_pd["newcol"] = newcol
    diamonds_dp = self.diamonds >> mutate(newcol=newcol)
    self.assertTrue(diamonds_pd.equals(diamonds_dp))

  def test_dupcolumn(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd["copy_x"] = diamonds_pd["x"]
    diamonds_dp = self.diamonds >> mutate(copy_x=X.x)
    self.assertTrue(diamonds_pd.equals(diamonds_dp))

  def test_editcolumn(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd["copy_x"] = 2 * diamonds_pd["x"]
    diamonds_dp = self.diamonds >> mutate(copy_x=X.x * 2)
    self.assertTrue(diamonds_pd.equals(diamonds_dp))

  def test_multi(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd["copy_x"] = diamonds_pd["x"]
    diamonds_pd["copy_y"] = diamonds_pd["y"]
    diamonds_dp = self.diamonds >> mutate(copy_x=X.x, copy_y=X.y)
    # Keep the new DplyFrame columns in the original order
    diamonds_dp = diamonds_dp[diamonds_pd.columns]
    self.assertTrue(diamonds_pd.equals(diamonds_dp))    

  def test_combine(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd["copy_x"] = diamonds_pd["x"] + diamonds_pd["y"]
    diamonds_dp = self.diamonds >> mutate(copy_x=X.x + X.y)
    self.assertTrue(diamonds_pd.equals(diamonds_dp))    

  def test_orignalUnaffect(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_dp = self.diamonds >> mutate(copy_x=X.x, copy_y=X.y)
    self.assertTrue(diamonds_pd.equals(self.diamonds))    

  def testCallMethodOnLater(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd["avgX"] = diamonds_pd.x.mean()
    diamonds_dp = self.diamonds >> mutate(avgX=X.x.mean())
    self.assertTrue(diamonds_pd.equals(diamonds_dp))    

  def testCallMethodOnCombinedLater(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd["avgX"] = (diamonds_pd.x + diamonds_pd.y).mean()
    diamonds_dp = self.diamonds >> mutate(avgX=(X.x + X.y).mean())
    self.assertTrue(diamonds_pd.equals(diamonds_dp))    

  def testReverseThings(self):
    dp = self.diamonds >> mutate(one_minus_carat=1 - X.carat, 
                                 carat_minus_one=X.carat-1)
    self.assertTrue(dp.one_minus_carat.equals(1 - self.diamonds.carat))
    self.assertTrue(dp.carat_minus_one.equals(self.diamonds.carat - 1))

  def testMethodFirst(self):
    diamonds_dp = self.diamonds >> mutate(avgDiff=X.x.mean() - X.x)
    diamonds_pd = self.diamonds.copy()
    diamonds_pd["avgDiff"] = diamonds_pd["x"].mean() - diamonds_pd["x"]
    self.assertTrue(diamonds_dp["avgDiff"].equals(diamonds_pd["avgDiff"]))

  def testArgsNotKwargs(self):
    diamonds_dp = mutate(self.diamonds, X.carat+1)
    diamonds_pd = self.diamonds.copy()
    diamonds_pd['X.carat + 1'] = diamonds_pd.carat + 1
    self.assertTrue(diamonds_pd.equals(diamonds_dp))

  def testOrderedKwargs(self):
    # without __order, is alphabetical
    diamonds_dp = mutate(self.diamonds,
                         carat2=X.carat+2,
                         carat1=X.carat+1,
                         carat3=X.carat+3)
    col_names = diamonds_dp.columns.values
    self.assertEqual(col_names[-3], "carat1")
    self.assertEqual(col_names[-2], "carat2")
    self.assertEqual(col_names[-1], "carat3")

    diamonds_dp = mutate(self.diamonds,
                         carat2=X.carat+2,
                         carat1=X.carat2-1,
                         carat3=X.carat1+2,
                         __order=["carat2", "carat1", "carat3"])
    
    col_names = diamonds_dp.columns.values
    self.assertEqual(col_names[-3], "carat2")
    self.assertEqual(col_names[-2], "carat1")
    self.assertEqual(col_names[-1], "carat3")
    self.assertTrue((diamonds_dp.carat + 3 == diamonds_dp.carat3).all())

  def testOrderedKwargsError(self):
  	self.assertRaisesRegexp(ValueError, "carat2", mutate,
  							self.diamonds, carat1 = X.carat + 1,
  							__order = ["carat1", "carat2"])

  	self.assertRaisesRegexp(ValueError, "carat3", mutate,
  							self.diamonds, carat1 = X.carat + 1, carat3 = X.carat + 3,
  							__order = ["carat1"])


class TestSelects(unittest.TestCase):
  diamonds = load_diamonds()

  def testOne(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd = diamonds_pd[["cut"]]
    diamonds_dp = self.diamonds >> select(X.cut)
    self.assertTrue(diamonds_pd.equals(diamonds_dp))    

  def testTwo(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd = diamonds_pd[["cut", "carat"]]
    diamonds_dp = self.diamonds >> select(X.cut, X.carat)
    self.assertTrue(diamonds_pd.equals(diamonds_dp))    

  def testChangeOrder(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd = diamonds_pd[["carat", "cut"]]
    diamonds_dp = self.diamonds >> select(X.carat, X.cut)
    self.assertTrue(diamonds_pd.equals(diamonds_dp))    

  def testGroupingImpliedSelect(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_grouped_1 = diamonds_pd >> group_by(X.cut, X.clarity) >> select(X.x)
    diamonds_selected_1 = diamonds_pd >> select(X.cut, X.clarity, X.x)
    diamonds_grouped_2 = diamonds_pd >> group_by(X.cut, X.clarity) >> select(X.x, X.clarity)
    diamonds_selected_2 = diamonds_pd >> select(X.cut, X.x, X.clarity)
    self.assertTrue(diamonds_grouped_1.equals(diamonds_selected_1))
    self.assertTrue(diamonds_grouped_2.equals(diamonds_selected_2))

class TestFilters(unittest.TestCase):
  diamonds = load_diamonds()

  def testFilterEasy(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd = diamonds_pd[diamonds_pd.cut == "Ideal"]
    diamonds_dp = self.diamonds >> sift(X.cut == "Ideal")
    self.assertTrue(diamonds_pd.equals(diamonds_dp))

  def testFilterNone(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_dp = self.diamonds >> sift()
    self.assertTrue(diamonds_pd.equals(diamonds_dp))    

  def testFilterWithMultipleArgs(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd = diamonds_pd[(diamonds_pd.cut == "Ideal") & 
                              (diamonds_pd.carat > 3)]
    diamonds_dp = self.diamonds >> sift(X.cut == "Ideal", X.carat > 3)
    self.assertTrue(diamonds_pd.equals(diamonds_dp))    

  def testFilterWithAnd(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd = diamonds_pd[(diamonds_pd.cut == "Ideal") & 
                              (diamonds_pd.carat > 3)]
    diamonds_dp = self.diamonds >> sift((X.cut == "Ideal") & (X.carat > 3))
    self.assertTrue(diamonds_pd.equals(diamonds_dp))    

  def testFilterWithOr(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd = diamonds_pd[(diamonds_pd.cut == "Ideal") |
                              (diamonds_pd.carat > 3)]
    diamonds_dp = self.diamonds >> sift((X.cut == "Ideal") | (X.carat > 3))
    self.assertTrue(diamonds_pd.equals(diamonds_dp))    

  def testFilterMultipleLaterColumns(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd = diamonds_pd[diamonds_pd.carat > diamonds_pd.x - diamonds_pd.y]
    diamonds_dp = self.diamonds >> sift(X.carat > X.x - X.y)
    self.assertTrue(diamonds_pd.equals(diamonds_dp))    


# TODO: Add more multiphase tests


class TestGroupBy(unittest.TestCase):
  diamonds = load_diamonds()

  def testGroupbyDoesntDie(self):
    self.diamonds >> group_by(X.color)

  def testUngroupDoesntDie(self):
    self.diamonds >> ungroup()

  def testGroupbyAndUngroupDoesntDie(self):
    self.diamonds >> group_by(X.color) >> ungroup()

  def testOneGroupby(self):
    diamonds_pd = self.diamonds.copy()
    carats_pd = set(diamonds_pd[diamonds_pd.carat > 3.5].groupby('color').mean()["carat"])
    diamonds_dp = (self.diamonds >> 
                    sift(X.carat > 3.5) >>
                    group_by(X.color) >> 
                    mutate(caratMean=X.carat.mean()))
    carats_dp = set(diamonds_dp["caratMean"].values)
    self.assertEqual(carats_pd, carats_dp)

  def testTwoGroupby(self):
    diamonds_pd = self.diamonds.copy()
    carats_pd = set(diamonds_pd[diamonds_pd.carat > 3.5].groupby(["color", "cut"]).mean()["carat"])
    diamonds_dp = (self.diamonds >> 
                    sift(X.carat > 3.5) >>
                    group_by(X.color, X.cut) >> 
                    mutate(caratMean=X.carat.mean()))
    carats_dp = set(diamonds_dp["caratMean"].values)
    self.assertEqual(carats_pd, carats_dp)

  def testGroupThenFilterDoesntDie(self):
    diamonds_dp = (self.diamonds >> 
                    group_by(X.color) >> 
                    sift(X.carat > 3.5) >>
                    mutate(caratMean=X.carat.mean()))

  def testGroupThenFilterDoesntDie2(self):
    diamonds_dp = (self.diamonds >> 
                    group_by(X.color) >> 
                    sift(X.carat > 3.5, X.color != "I") >>
                    mutate(caratMean=X.carat.mean()))

  def testGroupUngroupSummarize(self):
    num_rows = (self.diamonds >> group_by(X.cut) >> ungroup() >> 
                      summarize(total=X.price.sum()) >> nrow())
    self.assertEqual(num_rows, 1)
    sum_price = self.diamonds.sum()["price"]
    sum_price_dp = (self.diamonds >> group_by(X.cut) >> ungroup() >> 
                      summarize(total=X.price.sum()) >> X.total[0])
    self.assertEqual(sum_price, sum_price_dp)

  def testDfRemainsGroupedAfterOperation(self):
    diamonds_dp = (self.diamonds >>
                    group_by(X.color) >>
                    mutate(caratMean1=X.carat.mean()) >>
                    mutate(caratMean2=X.carat.mean()))
    self.assertTrue(diamonds_dp["caratMean1"].equals(diamonds_dp["caratMean2"]))

  def testGroupByWithArgsAndKwargs(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd['appearance'] = diamonds_pd['color'] + diamonds_pd['clarity']
    diamonds_pd = diamonds_pd.groupby(['cut', 'appearance'])['price'].sum()
    diamonds_grouped = (
      self.diamonds >>
      group_by(X.cut, appearance=X.color + X.clarity) >>
      summarize(total_price=X.price.sum())
    )
    for row in diamonds_grouped.itertuples():
      self.assertEqual(row.total_price, diamonds_pd[row.cut][row.appearance])

  def testGroupyByWithKwargsOnly(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd['appearance'] = diamonds_pd['color'] + diamonds_pd['clarity']
    diamonds_pd = diamonds_pd.groupby(['appearance'])['price'].sum()
    diamonds_grouped = (
      self.diamonds >>
      group_by(appearance=X.color + X.clarity) >>
      summarize(total_price=X.price.sum())
    )
    for row in diamonds_grouped.itertuples():
      self.assertEqual(row.total_price, diamonds_pd[row.appearance])

  def testGroupByWithPositionalArg(self):
    diamonds_pd = self.diamonds.copy()
    diamonds_pd['appearance'] = diamonds_pd['color'] + diamonds_pd['clarity']
    diamonds_pd = diamonds_pd.groupby(['cut', 'appearance'])['price'].sum()
    diamonds_grouped = (
      self.diamonds >>
      group_by(X.cut, X.color + X.clarity) >>
      summarize(total_price=X.price.sum())
    )
    self.assertTrue((diamonds_grouped.total_price ==  diamonds_pd.values).all())

  def testGroupByIterable(self):
    diamonds_pd = self.diamonds.copy()
    random_labels = np.random.choice(['a', 'b'], len(self.diamonds))
    diamonds_pd['label'] = random_labels
    diamonds_pd = diamonds_pd.groupby(['label'])['price'].sum()
    diamonds_grouped = (
      self.diamonds >>
      group_by(label=random_labels) >>
      summarize(total_price=X.price.sum())
    )
    for row in diamonds_grouped.itertuples():
      self.assertEqual(row.total_price, diamonds_pd[row.label])

  def testGroupByMutate(self):
    df = self.diamonds >> mutate(bin=X["Unnamed: 0"] % 5000)
    gbinp = df.groupby("bin")
    df["foo2"] = (df >> group_by(X.bin) >> mutate(foo=X.x.mean() + X.y.mean()))["foo"]
    df["foo1"] = gbinp.x.transform('mean') + gbinp.y.transform('mean')
    npt.assert_allclose(df.foo1, df.foo2)
    # self.assertAlmostEqual(df.foo1, df.foo2)


class TestArrange(unittest.TestCase):
  diamonds = load_diamonds()

  def testArrangeDoesntDie(self):
    self.diamonds >> arrange(X.cut)

  def testArrangeThenSelect(self):
    self.diamonds >> arrange(X.color) >> select(X.color)

  def testMultipleSort(self):
    self.diamonds >> arrange(X.color, X.cut) >> select(X.color)

  def testArrangeSorts(self):
    sortedColor_pd = self.diamonds.copy().sort_values("color")["color"]
    sortedColor_dp = (self.diamonds >> arrange(X.color))["color"]
    self.assertTrue(sortedColor_pd.equals(sortedColor_dp))

  def testMultiArrangeSorts(self):
    sortedCarat_pd = self.diamonds.copy().sort_values(["color", "carat"])["carat"]
    sortedCarat_dp = (self.diamonds >> arrange(X.color, X.carat))["carat"]
    self.assertTrue(sortedCarat_pd.equals(sortedCarat_dp))

  def testArrangeDescending(self):
    sortedCarat_pd = self.diamonds.copy().sort_values("carat", ascending=False)
    sortedCarat_dp = self.diamonds >> arrange(-X.carat)
    self.assertTrue((sortedCarat_pd.carat == sortedCarat_dp.carat).all())

  def testArrangeByComputedLater(self):
    sortedDf = self.diamonds >> arrange((X.carat-3)**2)
    self.assertEqual((sortedDf.iloc[0].carat-3)**2,
                     min((self.diamonds.carat-3)**2))
    self.assertEqual((sortedDf.iloc[-1].carat-3)**2,
                     max((self.diamonds.carat-3)**2))


class TestCount(unittest.TestCase):
  diamonds = load_diamonds()

  def testCount(self):
    counts_pd = Counter(self.diamonds.cut)
    counts_dp = self.diamonds >> count(X.cut)

    counts_pd = sorted(list(counts_pd.values()))
    counts_dp = sorted(list(counts_dp.n))

    self.assertEqual(counts_pd, counts_dp)

  def testCountExpression(self):
    counts_pd = Counter(self.diamonds.carat // 1)
    counts_dp = self.diamonds >> count(X.carat // 1)

    counts_pd = sorted(list(counts_pd.values()))
    counts_dp = sorted(list(counts_dp.n))

    self.assertEqual(counts_pd, counts_dp)

  def testCountMulti(self):
    counts_pd = Counter(zip(self.diamonds["cut"], self.diamonds["color"]))
    counts_dp = self.diamonds >> count(X.cut, X.color)

    counts_pd = sorted(list(counts_pd.values()))
    counts_dp = sorted(list(counts_dp.n))

    self.assertEqual(counts_pd, counts_dp)


class TestSample(unittest.TestCase):
  diamonds = load_diamonds()

  def testSamplesDontDie(self):
    self.diamonds >> sample_n(5)
    self.diamonds >> sample_frac(0.5)

  def testSamplesGetsRightNumber(self):
    shouldBe5 = self.diamonds >> sample_n(5) >> X._.__len__()
    self.assertEqual(shouldBe5, 5)
    frac = len(self.diamonds) * 0.1
    shouldBeFrac = self.diamonds >> sample_frac(0.1) >> X._.__len__()
    self.assertEqual(shouldBeFrac, frac)

  def testSampleEqualsPandasSample(self):
    for i in [1, 10, 100, 1000]:
      shouldBeI = self.diamonds >> sample_n(i) >> X._.__len__()
      self.assertEqual(shouldBeI, i)
    for i in [.1, .01, .001]:
      shouldBeI = self.diamonds >> sample_frac(i) >> X._.__len__()
      self.assertEqual(shouldBeI, round(len(self.diamonds)*i))

  def testSample0(self):
    shouldBe0 = self.diamonds >> sample_n(0) >> X._.__len__()
    self.assertEqual(shouldBe0, 0)
    shouldBeFrac = self.diamonds >> sample_frac(0) >> X._.__len__()
    self.assertEqual(shouldBeFrac, 0.)

  def testGroupedSample(self):
    num_groups = len(set(self.diamonds["cut"]))
    for i in [0, 1, 10, 100, 1000]:
      numRows = self.diamonds >> group_by(X.cut) >> sample_n(i) >> X._.__len__()
      self.assertEqual(numRows, i*num_groups)
    for i in [.1, .01, .001]:
      shouldBeI = self.diamonds >> group_by(X.cut) >> sample_frac(i) >> X._.__len__()
      out = sum([len(self.diamonds[self.diamonds.cut == c].sample(frac=i)) for c in set(self.diamonds.cut)])
      # self.assertEqual(shouldBeI, math.floor(len(self.diamonds)*i))
      self.assertEqual(shouldBeI, out)


class TestSummarize(unittest.TestCase):
  diamonds = load_diamonds()

  def testSummarizeDoesntDie(self):
    self.diamonds >> summarize(sumX=X.x.sum())

  def testSummarizeX(self):
    diamonds_pd = self.diamonds.copy()
    sumX_pd = diamonds_pd.sum()["x"]
    sumX_dp = (self.diamonds >> summarize(sumX=X.x.sum()))["sumX"][0]
    self.assertEqual(round(sumX_pd), round(sumX_dp))

  def testSummarizeGroupedX(self):
    diamonds_pd = self.diamonds.copy()
    sumX_pd = diamonds_pd.groupby("cut").sum()["x"]
    val_pd = sumX_pd.values.copy()
    val_pd.sort()
    valX_dp = (self.diamonds >> group_by(X.cut) >>
                summarize(sumX=X.x.sum()) >> X._["sumX"]).values.copy()
    valX_dp.sort()
    for i, j in zip(val_pd, valX_dp):
      self.assertEqual(round(i), round(j))

  def testSummarizeGrouping(self):
    a = DplyFrame(pd.DataFrame({'x': [1, 1, 2, 3]
                                , 'y': range(1, 5)}))
    a_summarized = a >> group_by(X.x) >> summarize(mean_y = X.y.mean())
    self.assertIsNone(a_summarized._grouped_on)


class TestAlternateAttrGrab(unittest.TestCase):
  diamonds = load_diamonds()
  diamonds["o m g"] = range(len(diamonds))
  diamonds["0"] = range(len(diamonds))

  def testSelect(self):
    equality = self.diamonds[["o m g"]] == (self.diamonds >> select(X["o m g"]))
    self.assertTrue(equality.all()[0])

  def testMutate(self):
    pd = self.diamonds[["0"]] * 2
    dp = self.diamonds >> mutate(foo=X["0"]*2) >> select(X.foo)
    equality = pd["0"] == dp["foo"]
    self.assertTrue(equality.all())


class TestNrow(unittest.TestCase):
  diamonds = load_diamonds()

  def testSimpleNrow(self):
    diamonds_pd = self.diamonds.copy()
    self.assertEqual(len(diamonds_pd), self.diamonds >> nrow())

  def testMultipleNrow(self):
    diamonds_pd = self.diamonds.copy()
    self.assertEqual(len(diamonds_pd), self.diamonds >> nrow())
    self.assertEqual(len(diamonds_pd), self.diamonds >> nrow())

    small_d = diamonds_pd[diamonds_pd.carat > 4]
    self.assertEqual(
        len(small_d), self.diamonds >> sift(X.carat > 4) >> nrow())


class TestFunctionForm(unittest.TestCase):
  diamonds = load_diamonds()

  def testsift(self):
    normal = self.diamonds >> sift(X.carat > 4)
    function = sift(self.diamonds, X.carat > 4)
    self.assertTrue(normal.equals(function))

    normal = self.diamonds >> sift()
    function = sift(self.diamonds)
    self.assertTrue(normal.equals(function))

    normal = self.diamonds >> sift(X.carat < 4, X.color == "D")
    function = sift(self.diamonds, X.carat < 4, X.color == "D")
    self.assertTrue(normal.equals(function))

    normal = self.diamonds >> sift((X.carat < 4) | (X.color == "D"), 
                                      X.carat >= 4)
    function = sift(self.diamonds, (X.carat < 4) | (X.color == "D"), 
                                      X.carat >= 4)
    self.assertTrue(normal.equals(function))

  def testSelect(self):
    normal = self.diamonds >> select(X.carat)
    function = select(self.diamonds, X.carat)
    self.assertTrue(normal.equals(function))

    normal = self.diamonds >> select(X.cut, X.carat, X.carat)
    function = select(self.diamonds, X.cut, X.carat, X.carat)
    self.assertTrue(normal.equals(function))

  def testMutate(self):
    normal = self.diamonds >> mutate(foo=X.carat)
    function = mutate(self.diamonds, foo=X.carat)
    self.assertTrue(normal.equals(function))

    normal = self.diamonds >> mutate(a=X.cut, b=X.x/2, c=32)
    function = mutate(self.diamonds, a=X.cut, b=X.x/2, c=32)
    self.assertTrue(normal.equals(function))

  def testGroupBy(self):
    normal = (self.diamonds >> 
                    sift(X.carat > 3.5) >>
                    group_by(X.color) >> 
                    mutate(caratMean=X.carat.mean()))
    function = self.diamonds >> sift(X.carat > 3.5)
    function = group_by(function, X.color)
    function = function >> mutate(caratMean=X.carat.mean())
    self.assertTrue(normal.equals(function))

  def testGroupBy2(self):
    normal = (self.diamonds >> 
                    group_by(X.color, X.cut) >> 
                    mutate(caratMean=X.carat.mean()))
    function = group_by(self.diamonds, X.color, X.cut)
    function = function >> mutate(caratMean=X.carat.mean())
    self.assertTrue(normal.equals(function))

  def testUngroup(self):
    normal = (self.diamonds >> 
                    sift(X.carat > 3.5) >>
                    group_by(X.color) >> 
                    ungroup())
    function = (self.diamonds >> 
                    sift(X.carat > 3.5) >>
                    group_by(X.color))
    function = ungroup(function)
    self.assertTrue(normal.equals(function))

  def testArrange(self):
    normal = self.diamonds >> arrange(X.color, X.carat)
    function = arrange(self.diamonds, X.color, X.carat)
    self.assertTrue(normal.equals(function))

    normal = self.diamonds >> arrange(X.cut, X.carat, X.y)
    function = arrange(self.diamonds, X.cut, X.carat, X.y)
    self.assertTrue(normal.equals(function))

  def testSummarize(self):
    normal = (self.diamonds >> summarize(sumX=X.x.sum()))
    function = summarize(self.diamonds, sumX=X.x.sum())
    self.assertTrue(normal.equals(function))

  def testSummarizeGroupedX(self):
    normal = self.diamonds >> group_by(X.cut) >> summarize(sumX=X.x.sum())
    function = self.diamonds >> group_by(X.cut)
    function = summarize(function, sumX=X.x.sum())
    self.assertTrue(normal.equals(function))

  def testUtilities(self):
    normal = self.diamonds >> head()
    function = head(self.diamonds)
    self.assertTrue(normal.equals(function))

    normal = self.diamonds >> sample_n(10)
    function = sample_n(self.diamonds, 10)
    self.assertEqual(len(normal), len(function))
    
    normal = self.diamonds >> sample_frac(0.1)
    function = sample_frac(self.diamonds, 0.1)
    self.assertEqual(len(normal), len(function))
    

class TestIfElse(unittest.TestCase):
  diamonds = load_diamonds()

  def test_if_else(self):
    foo = self.diamonds >> mutate(
        conditional_results= if_else(X.cut == "Premium", X.color, X.clarity))
    bar = [color if cut == "Premium" else clarity for color, clarity, cut
        in zip(self.diamonds.color, self.diamonds.clarity, self.diamonds.cut)]
    bar = pd.Series(bar)
    self.assertTrue(foo["conditional_results"].equals(bar))

  # Porting dplyr tests:
  # https://github.com/hadley/dplyr/blob/master/tests/testthat/test-if-else.R
  def test_if_else_work(self):
    x = pd.Series([-1, 0, 1])
    zeros = pd.Series([0, 0, 0])
    self.assertTrue(if_else(x < 0, x, zeros).equals(pd.Series([-1, 0, 0])))
    self.assertTrue(if_else(x > 0, x, zeros).equals(pd.Series([0, 0, 1])))


class TestRename(unittest.TestCase):
  diamonds = load_diamonds()

  def test_rename(self):
    renamed_df = self.diamonds >> rename(chair=X.table)
    self.assertNotIn('table', renamed_df.columns)
    self.assertIn('chair', renamed_df.columns)


class TestTransmute(unittest.TestCase):
  diamonds = load_diamonds()

  def test_transmute(self):
    mutate_select_df = (self.diamonds >>
                        mutate(new_price=X.price * 2, x_plus_y=X.x + X.y,
                               __order=['new_price', 'x_plus_y']) >>
                        select(X.new_price, X.x_plus_y))
    transmute_df = (self.diamonds >>
                    transmute(new_price=X.price * 2, x_plus_y=X.x + X.y,
                              __order=('new_price', 'x_plus_y')))
    self.assertTrue(mutate_select_df.equals(transmute_df))

  def test_transmute_args(self):
    mutate_select_df = (self.diamonds >>
                        mutate(X.price * 2, X.x + X.y) >>
                        select(X["X.price * 2"], X["X.x + X.y"]))
    transmute_df = (self.diamonds >>
                    transmute(X.price * 2, X.x + X.y))
    self.assertTrue(mutate_select_df.equals(transmute_df))


class TestMutatingJoins(unittest.TestCase):

  def test_inner_join(self):
    a = DplyFrame(pd.DataFrame({'x': [1, 1, 2, 3]
                              , 'y': [1, 2, 3, 4]}))
    b = DplyFrame(pd.DataFrame({'x': [1, 2, 2, 4]
                              , 'z': [1, 2, 3, 4]}))
    j_inner_test = a >> inner_join(b, by=['x'])
    j_inner_pd = DplyFrame(pd.DataFrame({'x': [1, 1, 2, 2]
                       , 'y': [1, 2, 3, 3]
                       , 'z': [1, 1, 2, 3]}))
    self.assertTrue(j_inner_test.equals(j_inner_pd))
    self.assertTrue(len(j_inner_test.columns.difference(j_inner_pd.columns)) == 0)
    # test normal function
    j1 = inner_join(a, b)
    j2 = a >> inner_join(b)
    self.assertTrue(j1.equals(j2))
    j1 = inner_join(a, b, by=['x'])
    j2 = a >> inner_join(b, by=['x'])
    self.assertTrue(j1.equals(j2))
    # test on grouped data
    j1 = a >> group_by(X.x) >> inner_join(b)
    # dataframes compare equal...
    self.assertTrue(j1.equals(j2))
    # but have different grouping
    self.assertTrue(j1._grouped_on == ['x'])
    self.assertTrue(j2._grouped_on is None)



  def test_full_join(self):
    a = DplyFrame(pd.DataFrame({'x': [1, 2, 3]
                                , 'y': [2, 3, 4]}))
    b = DplyFrame(pd.DataFrame({'x': [3, 4, 5]
                                , 'z': [3, 4, 5]}))
    j_full_test = a >> full_join(b, by=['x'])
    j_full_pd = DplyFrame(pd.DataFrame({'x': [1.0, 2.0, 3.0, 4.0, 5.0] # pandas promotes ints to floats in the join
                              , 'y': [2, 3, 4, np.nan, np.nan]
                              , 'z': [np.nan, np.nan, 3, 4, 5]}))
    self.assertTrue(j_full_test.equals(j_full_pd))
    self.assertTrue(len(j_full_test.columns.difference(j_full_pd.columns)) == 0)
    # test normal form
    j1 = full_join(a, b)
    j2 = a >> full_join(b)
    self.assertTrue(j1.equals(j2))
    j1 = full_join(a, b, by=['x'])
    j2 = a >> full_join(b, by=['x'])
    self.assertTrue(j1.equals(j2))
    # test on grouped data
    j1 = a >> group_by(X.x) >> full_join(b)
    # dataframes compare equal...
    self.assertTrue(j1.equals(j2))
    # but have different grouping
    self.assertTrue(j1._grouped_on == ['x'])
    self.assertTrue(j2._grouped_on is None)


  def test_left_join(self):
    a = DplyFrame(pd.DataFrame({'x': [1, 1, 2, 3]
                                , 'y': [1, 2, 3, 4]}))
    b = DplyFrame(pd.DataFrame({'x': [1, 2, 2, 4]
                                , 'z': [1, 2, 3, 4]}))
    j_left_test = a >> left_join(b, by=['x'])
    j_left_pd = DplyFrame(pd.DataFrame({'x': [1, 1, 2, 2, 3]
                              , 'y': [1, 2, 3, 3, 4]
                              , 'z': [1, 1, 2, 3, np.nan]}))
    j_left_test.equals(j_left_pd)
    self.assertTrue(len(j_left_test.columns.difference(j_left_pd.columns)) == 0)
    # test normal
    j1 = left_join(a, b)
    j2 = a >> left_join(b)
    self.assertTrue(j1.equals(j2))
    j1 = left_join(a, b, by=['x'])
    j2 = a >> left_join(b, by=['x'])
    self.assertTrue(j1.equals(j2))
    # test on grouped data
    j1 = a >> group_by(X.x) >> left_join(b)
    # dataframes compare equal...
    self.assertTrue(j1.equals(j2))
    # but have different grouping
    self.assertTrue(j1._grouped_on == ['x'])
    self.assertTrue(j2._grouped_on is None)

  def test_right_join(self):
    a = DplyFrame(pd.DataFrame({'x': [1, 1, 2, 3]
                                , 'y': [1, 2, 3, 4]}))
    b = DplyFrame(pd.DataFrame({'x': [1, 2, 2, 4]
                                , 'z': [1, 2, 3, 4]}))
    j_right_test = a >> right_join(b, by=['x'])
    j_right_pd = DplyFrame(pd.DataFrame({'x': [1.0, 1.0, 2.0, 2.0, 4.0]
                              , 'y': [1.0, 2.0, 3.0, 3.0, np.nan]
                              , 'z': [1, 1, 2, 3, 4]}))
    self.assertTrue(j_right_test.equals(j_right_pd))
    self.assertTrue(len(j_right_test.columns.difference(j_right_pd.columns)) == 0)
    # test normal form
    j1 = right_join(a, b)
    j2 = a >> right_join(b)
    self.assertTrue(j1.equals(j2))
    j1 = right_join(a, b, by=['x'])
    j2 = a >> right_join(b, by=['x'])
    self.assertTrue(j1.equals(j2))
    # test on grouped data
    j1 = a >> group_by(X.x) >> right_join(b)
    # dataframes compare equal...
    self.assertTrue(j1.equals(j2))
    # but have different grouping
    self.assertTrue(j1._grouped_on == ['x'])
    self.assertTrue(j2._grouped_on is None)

  def test_suffixes_join(self):
    a = DplyFrame(pd.DataFrame({'x': [1, 1, 2, 3]
                                , 'z': [1, 2, 3, 4]}))
    b = DplyFrame(pd.DataFrame({'x': [1, 2, 2, 4]
                                , 'z': [1, 2, 3, 4]}))
    j_suffix_test = a >> left_join(b, by=['x'])
    j_suffix_pd = DplyFrame(pd.DataFrame({'x': [1, 1, 2, 2, 3]
                              , 'z_x': [1, 2, 3, 3, 4]
                              , 'z_y': [1.0, 1.0, 2.0, 3.0, np.nan]}))
    self.assertTrue((j_suffix_test.columns == j_suffix_pd.columns).all())
    j_suffix_test = a >> left_join(b, by=['x'], suffixes=('_1', '_2'))
    j_suffix_pd = DplyFrame(pd.DataFrame({'x': [1, 1, 2, 2, 3]
                              , 'z_1': [1, 2, 3, 3, 4]
                              , 'z_2': [1.0, 1.0, 2.0, 3.0, np.nan]}))
    self.assertTrue((j_suffix_test.columns == j_suffix_pd.columns).all())

  def test_multiple_columns_join(self):

    a = DplyFrame(pd.DataFrame({'x': [1, 1, 2, 3]
                                , 'y': [1, 1, 2, 3]
                                , 'a': [1, 2, 3, 4]})[['x', 'y', 'a']])
    b = DplyFrame(pd.DataFrame({'x': [1, 2, 2, 4]
                                , 'y': [1, 2, 2, 4]
                                , 'b': [1, 2, 3, 4]})[['x', 'y', 'b']])
    j_multiple_col_test = a >> left_join(b, by=['x', 'y'])
    j_multiple_col_pd = DplyFrame(pd.DataFrame({'x': [1, 1, 2, 2, 3]
                                      , 'y': [1, 1, 2, 2, 3]
                                      , 'a': [1, 2, 3, 3, 4]
                                      , 'b': [1.0, 1.0, 2.0, 3.0, np.nan]}))[['x', 'y', 'a', 'b']]
    self.assertTrue(j_multiple_col_test.equals(j_multiple_col_pd))
    b = DplyFrame(pd.DataFrame({'z': [1, 2, 2, 4]
                                , 'y': [1, 2, 2, 4]
                                , 'b': [1, 2, 3, 4]})[['z', 'y', 'b']])
    j_multiple_col_test = a >> left_join(b, by=[('x', 'z'), 'y'])
    j_multiple_col_pd =DplyFrame(pd.DataFrame({'x': [1, 1, 2, 2, 3]
                                      , 'y': [1, 1, 2, 2, 3]
                                      , 'a': [1, 2, 3, 3, 4]
                                      , 'z': [1.0, 1.0, 2.0, 2.0, np.nan]
                                      , 'b': [1.0, 1.0, 2.0, 3.0, np.nan]}))[['x', 'y', 'a', 'z', 'b']]
    self.assertTrue(j_multiple_col_test.equals(j_multiple_col_pd))


class TestFilteringJoins(unittest.TestCase):

  superheroes = """name,alignment,gender,publisher
Magneto,bad,male,Marvel
Storm,good,female,Marvel
Mystique,bad,female,Marvel
Batman,good,male,DC
Joker,bad,male,DC
Catwoman,bad,female,DC
Hellboy,good,male,DarkHorseComics
"""
  superheroes = DplyFrame(pd.read_csv(StringIO(superheroes)))

  publishers = """publisher,yr_founded
DC,1934
Marvel,1939
Image,1992
"""
  publishers = DplyFrame(pd.read_csv(StringIO(publishers)))
  publishers_2 = publishers.rename(columns={'publisher': 'publisher_2', 'yr_founded': 'yr_founded_2'})

  def test_semi_join_1(self):
    j_test = self.superheroes >> semi_join(self.publishers)
    j_pd = DplyFrame(pd.read_csv(StringIO("""name,alignment,gender,publisher
Magneto,bad,male,Marvel
Storm,good,female,Marvel
Mystique,bad,female,Marvel
Batman,good,male,DC
Joker,bad,male,DC
Catwoman,bad,female,DC""")))
    self.assertTrue(j_test.equals(j_pd))
    # names don't matter
    j_test = self.superheroes >> semi_join(self.publishers_2, by=[('publisher', 'publisher_2')])
    self.assertTrue(j_test.equals(j_pd))
    # works in normal form
    j1 = semi_join(self.superheroes, self.publishers)
    j2 = self.superheroes >> semi_join(self.publishers)
    self.assertTrue(j1.equals(j2))
    # works on grouped data
    j1 = self.superheroes >> group_by(X.publisher) >> semi_join(self.publishers)
    # compares equal
    self.assertTrue(j1.equals(j2))
    # but don't group the same
    self.assertTrue(j1._grouped_on == ['publisher'])
    self.assertTrue(j2._grouped_on is None)


  def test_anti_join_1(self):
    j_test = self.superheroes >> anti_join(self.publishers)
    j_pd = DplyFrame(pd.read_csv(StringIO("""index,name,alignment,gender,publisher
6,Hellboy,good,male,DarkHorseComics""")).set_index(['index']))
    j_pd.index.name = None
    self.assertTrue(j_test.equals(j_pd))
    j_test = self.superheroes >> anti_join(self.publishers_2, by=[('publisher', 'publisher_2')])
    # names don't matter
    self.assertTrue(j_test.equals(j_pd))
    # works in normal form
    j1 = anti_join(self.superheroes, self.publishers)
    j2 = self.superheroes >> anti_join(self.publishers)
    self.assertTrue(j1.equals(j2))
    # works on grouped data
    j1 = self.superheroes >> group_by(X.publisher) >> anti_join(self.publishers)
    # compares equal
    self.assertTrue(j1.equals(j2))
    # but don't group the same
    self.assertTrue(j1._grouped_on == ['publisher'])
    self.assertTrue(j2._grouped_on is None)

  def test_semi_join_2(self):
    j_test = self.publishers >> semi_join(self.superheroes)
    j_pd = DplyFrame(pd.read_csv(StringIO("""publisher,yr_founded
DC,1934
Marvel,1939""")))
    self.assertTrue(j_test.equals(j_pd))
    # names don't matter
    j_test = self.publishers_2 >> semi_join(self.superheroes, by=[('publisher_2', 'publisher')])
    j_pd = DplyFrame(pd.read_csv(StringIO("""publisher_2,yr_founded_2
DC,1934
Marvel,1939""")))
    self.assertTrue(j_test.equals(j_pd))

  def test_anti_join_2(self):
    j_test = self.publishers >> anti_join(self.superheroes)
    j_pd = DplyFrame(pd.read_csv(StringIO("""index,publisher,yr_founded
2,Image,1992""")).set_index(['index']))
    self.assertTrue(j_test.equals(j_pd))
    # names don't matter
    j_test = self.publishers_2 >> anti_join(self.superheroes, by=[('publisher_2', 'publisher')])
    j_pd = DplyFrame(pd.read_csv(StringIO("""index,publisher_2,yr_founded_2
2,Image,1992""")).set_index(['index']))
    self.assertTrue(j_test.equals(j_pd))

  a = DplyFrame(pd.read_csv(StringIO("""x,y
1,1
1,2
2,3
3,4""")))
  b = DplyFrame(pd.read_csv(StringIO("""x,z
1,1
2,2
2,3
4,4""")))

  def test_semi_join_dplyr_1(self):
    j_test_1 = self.a >> semi_join(self.b)
    j_test_2 = self.b >> semi_join(self.a)
    j_pd_1 = DplyFrame(pd.read_csv(StringIO("""x,y
1,1
1,2
2,3""")))
    j_pd_2 = DplyFrame(pd.read_csv(StringIO("""x,z
1,1
2,2
2,3""")))
    self.assertTrue(j_test_1.equals(j_pd_1))
    self.assertTrue(j_test_2.equals(j_pd_2))

  def test_anti_join_dplyr_1(self):
    j_test_1 = self.a >> anti_join(self.b)
    j_test_2 = self.b >> anti_join(self.a)
    j_pd_1 = DplyFrame(pd.read_csv(StringIO("""index,x,y
3,3,4""")).set_index(['index']))
    j_pd_1.index.name = None
    j_pd_2 = DplyFrame(pd.read_csv(StringIO("""index,x,z
3,4,4""")).set_index(['index']))
    j_pd_2.index.name = None
    self.assertTrue(j_test_1.equals(j_pd_1))
    self.assertTrue(j_test_2.equals(j_pd_2))

  c = DplyFrame(pd.read_csv(StringIO("""x,y,a
1,1,1
1,1,2
2,2,3
3,3,4""")))
  d = DplyFrame(pd.read_csv(StringIO("""x,y,b
1,1,1
2,2,2
2,2,3
4,4,4""")))

  def test_semi_join_dplyr_2(self):
    # bivariate keys
    j_test_1 = self.c >> semi_join(self.d)
    j_test_2 = self.d >> semi_join(self.c)
    j_pd_1 = DplyFrame(pd.read_csv(StringIO("""x,y,a
1,1,1
1,1,2
2,2,3""")))
    j_pd_2 = DplyFrame(pd.read_csv(StringIO("""x,y,b
1,1,1
2,2,2
2,2,3""")))
    self.assertTrue(j_test_1.equals(j_pd_1))
    self.assertTrue(j_test_2.equals(j_pd_2))
    # include column names
    j_test_1 = self.c >> semi_join(self.d, by=['x', 'y'])
    j_test_2 = self.d >> semi_join(self.c, by=['x', 'y'])
    self.assertTrue(j_test_1.equals(j_pd_1))
    self.assertTrue(j_test_2.equals(j_pd_2))
    # use different column names
    alt_c = self.c.rename(columns={'x': 'x_2'})
    j_test_1 = alt_c >> semi_join(self.d, by=[('x_2', 'x'), 'y'])
    j_test_2 = self.d >> semi_join(alt_c, by=[('x', 'x_2'), 'y'])
    j_pd_1 = DplyFrame(pd.read_csv(StringIO("""x_2,y,a
1,1,1
1,1,2
2,2,3""")))
    self.assertTrue(j_test_1.equals(j_pd_1))
    self.assertTrue(j_test_2.equals(j_pd_2))

  def test_anti_join_dplyr_2(self):
    # bivariate keys
    j_test_1 = self.c >> anti_join(self.d)
    j_test_2 = self.d >> anti_join(self.c)
    j_pd_1 = DplyFrame(pd.read_csv(StringIO("""index,x,y,a
3,3,3,4""")).set_index(['index']))
    j_pd_1.index.name = None
    j_pd_2 = DplyFrame(pd.read_csv(StringIO("""index,x,y,b
3,4,4,4""")).set_index(['index']))
    j_pd_2.index.name = None
    self.assertTrue(j_test_1.equals(j_pd_1))
    self.assertTrue(j_test_2.equals(j_pd_2))
    # use column names
    j_test_1 = self.c >> anti_join(self.d, by=['x', 'y'])
    j_test_2 = self.d >> anti_join(self.c, by=['x', 'y'])
    self.assertTrue(j_test_1.equals(j_pd_1))
    self.assertTrue(j_test_2.equals(j_pd_2))
    # use different column names
    alt_c = self.c.rename(columns={'x': 'x_2'})
    j_test_1 = alt_c >> anti_join(self.d, by=[('x_2', 'x'), 'y'])
    j_test_2 = self.d >> anti_join(alt_c, by=[('x', 'x_2'), 'y'])
    j_pd_1 = DplyFrame(pd.read_csv(StringIO("""index,x_2,y,a
3,3,3,4""")).set_index(['index']))
    j_pd_1.index.name = None
    self.assertTrue(j_test_1.equals(j_pd_1))
    self.assertTrue(j_test_2.equals(j_pd_2))



if __name__ == '__main__':
  unittest.main()




from dplython import X, mutate, group_by, diamonds
diamonds = diamonds >> mutate(bin=X["Unnamed: 0"] % 5000)
gbinp = diamonds.groupby("bin")
gbind = diamonds >> group_by(X.bin)

# Test 1
gbinp["foo"] = gbinp.x.transform('mean')
gbind = gbind >> mutate(foo=X.x.mean())
print gbinp["foo"].equals(gbind["foo"])

# Test 2
gbinp["foo"] = gbinp.x.transform('mean') + gbinp.y.transform('mean')
gbind = gbind >> mutate(foo=X.x.mean() + X.y.mean())
print gbinp["foo"].equals(gbind["foo"])




"""Install Dplython."""


from setuptools import setup, find_packages


setup(
    name="dplython",
    version="0.0.7",
    description="Dplyr-style operations on top of pandas DataFrame.",
    url="https://github.com/dodger487/dplython",
    download_url="https://github.com/dodger487/dplython/tarball/0.0.7",
    packages=find_packages(),
    license="MIT",
    keywords="pandas data dplyr",
    package_data={"dplython": ["data/diamonds.csv"]},
    package_dir={"dplython": "dplython"},
    install_requires=["numpy", "pandas", "six"],
    author="Chris Riederer",
    author_email="OfficialChrisEmail@gmail.com",
    maintainer="Chris Riederer",
    maintainer_email="OfficialChrisEmail@gmail.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Environment :: Web Environment",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Scientific/Engineering",
    ]
)


# Chris Riederer
# 2016-02-17

"""Dplyr-style operations on top of pandas DataFrame."""

from functools import wraps
import itertools
import sys
import types
import warnings
warnings.simplefilter("once")

import six
from six.moves import range

import numpy as np
import pandas
from pandas import DataFrame

try:
  from .later import (Later, CreateLaterFunction, DelayFunction, X, Manager)
except:
  from later import (Later, CreateLaterFunction, DelayFunction, X, Manager)


__version__ = "0.0.7"


def _addQuotes(item):
  return '"' + item + '"' if isinstance(item, str) else item


class DplyFrame(DataFrame):
  """A subclass of the pandas DataFrame with methods for function piping.

  This class implements two main features on top of the pandas DataFrame. First,
  dplyr-style groups. In contrast to SQL-style or pandas style groups, rows are 
  not collapsed and replaced with a function value.
  Second, >> is overloaded on the DataFrame so that functions on the right-hand
  side of this equation are called on the object. For example,
  
  >>> df >> select(X.carat)
  
  will call a function (created from the "select" call) on df.

  Currently, these inputs need to be one of the following:

  * A "Later" 
  * The "ungroup" function call
  * A function that returns a pandas DataFrame or DplyFrame.
  """
  _metadata = ["_grouped_on", "_grouped_self"]

  def __init__(self, *args, **kwargs):
    super(DplyFrame, self).__init__(*args, **kwargs)
    self._grouped_on = None
    self._current_group = None
    self._grouped_self = None
    if len(args) == 1 and isinstance(args[0], DplyFrame):
      self._copy_attrs(args[0])

  def _copy_attrs(self, df):
    for attr in self._metadata:
      self.__dict__[attr] = getattr(df, attr, None)

  @property
  def _constructor(self):
    return DplyFrame

  def group_self(self, names):
    self._grouped_on = names
    self._grouped_self = self.groupby(names)

  def ungroup(self):
    self._grouped_on = None
    self._grouped_self = None

  def regroup(self, names):
    self.group_self(names)
    return self

  def apply_on_groups(self, delayedFcn):
    handled_classes = (mutate, sift, inner_join, full_join, left_join, 
                       right_join, semi_join, anti_join, summarize)
    if isinstance(delayedFcn, handled_classes):
      return delayedFcn(self)

    outDf = self._grouped_self.apply(delayedFcn)

    # Remove multi-index created from grouping and applying
    for grouped_name in outDf.index.names[:-1]:
      if grouped_name in outDf:
        outDf.reset_index(level=0, drop=True, inplace=True)
      else:
        outDf.reset_index(level=0, inplace=True)

    # Drop all 0 index, created by summarize
    if (outDf.index == 0).all():
      outDf.reset_index(drop=True, inplace=True)
    outDf.group_self(self._grouped_on)
    return outDf

  def __rshift__(self, delayedFcn):
    if type(delayedFcn) == Later:
      return delayedFcn.evaluate(self)

    if delayedFcn == UngroupDF:
      otherDf = DplyFrame(self.copy(deep=True))
      return delayedFcn(otherDf)

    if self._grouped_self:
      outDf = self.apply_on_groups(delayedFcn)
      return outDf
    else:
      otherDf = DplyFrame(self.copy(deep=True))
      return delayedFcn(otherDf)


def ApplyToDataframe(fcn):
  @wraps(fcn)
  def DplyrFcn(*args, **kwargs):
    data_arg = None
    if len(args) > 0 and isinstance(args[0], pandas.DataFrame):
      # data_arg = args[0].copy(deep=True)
      data_arg = args[0]
      if 'join' not in fcn.__name__:
        args = args[1:]
    fcn_to_apply = fcn(*args, **kwargs)
    if data_arg is None:
      return fcn_to_apply
    else:
      return data_arg >> fcn_to_apply
  return DplyrFcn


class Verb(object):

  def __new__(cls, *args, **kwargs):
    if len(args) > 0 and isinstance(args[0], pandas.DataFrame):
      verb = cls(*args[1:], **kwargs)
      return verb(args[0].copy(deep=True))
    else:
      return super(Verb, cls).__new__(cls)

  def __init__(self, *args, **kwargs):
    self.args = args
    self.kwargs = kwargs

  def do(self):
    raise NotImplementedError()


class sift(Verb):
  """Filters rows of the data that meet input criteria.

  Giving multiple arguments to sift is equivalent to a logical "and".
  
  >>> df >> sift(X.carat > 4, X.cut == "Premium")
  # Out:
  # carat      cut color clarity  depth  table  price      x  ...
  #  4.01  Premium     I      I1   61.0     61  15223  10.14
  #  4.01  Premium     J      I1   62.5     62  15223  10.02
  
  As in pandas, use bitwise logical operators like ``|``, ``&``:
  
  >>> df >> sift((X.carat > 4) | (X.cut == "Ideal")) >> head(2)
  # Out:  carat    cut color clarity  depth ...
  #        0.23  Ideal     E     SI2   61.5     
  #        0.23  Ideal     J     VS1   62.8     
  """

  __name__ = "sift"

  def __call__(self, df):
    # TODO: This function is a candidate for improvement!
    final_filter = pandas.Series([True for t in range(len(df))])
    final_filter.index = df.index
    grouped = "transform" if df._grouped_on else None
    for arg in self.args:
      stmt = arg.evaluate(df, special=grouped)
      final_filter = final_filter & stmt
    if final_filter.dtype != bool:
      raise Exception("Inputs to filter must be boolean")
    df = df[final_filter]
    if grouped:
      df.group_self(df._grouped_on)
    return df

  def __rrshift__(self, other):
    return self.__call__(other)



def dfilter(*args, **kwargs):
  warnings.warn("'dfilter' is deprecated. Please use 'sift' instead.",
                DeprecationWarning)
  return sift(*args, **kwargs)


@ApplyToDataframe
def select(*args):
  """Select specific columns from DataFrame. 

  Output will be DplyFrame type. Order of columns will be the same as input into
  select.

  >>> diamonds >> select(X.color, X.carat) >> head(3)
  Out:
    color  carat
  0     E   0.23
  1     E   0.21
  2     E   0.23

  Grouping variables are implied in selection.
  >>> df >> group_by(X.a, X.b) >> select(X.c)
  returns a dataframe like `df[[X.a, X.b, X.c]]` with the variables appearing in
  grouped order before the selected column(s), unless a grouped variable is
  explicitly selected

  >>> df >> group_by(X.a, X.b) >> select(X.c, X.b)
  returns a dataframe like `df[[X.a, X.c, X.b]]`
  """
  def select_columns(df, args):
    columns = [column._name for column in args]
    if df._grouped_on:
      for col in df._grouped_on[::-1]:
        if col not in columns:
          columns.insert(0, col)
    return columns
  return lambda df: df[select_columns(df, args)]


def _dict_to_possibly_ordered_tuples(dict_):
  order = dict_.pop("__order", None)
  if order:
    ordered_keys = set(order)
    dict_keys = set(dict_)

    missing_order = ordered_keys - dict_keys
    if missing_order:
      raise ValueError(", ".join(missing_order) +
                       " in __order not found in keyword arguments")

    missing_kwargs = dict_keys - ordered_keys
    if missing_kwargs:
      raise ValueError(", ".join(missing_kwargs) + " not found in __order")

    return [(key, dict_[key]) for key in order]
  else:
    return sorted(dict_.items(), key=lambda e: e[0])


class mutate(Verb):
  """Adds a column to the DataFrame.

  This can use existing columns of the DataFrame as input.

  >>> (diamonds >>
          mutate(carat_bin=X.carat.round()) >>
          group_by(X.cut, X.carat_bin) >>
          summarize(avg_price=X.price.mean()))
  Out:
         avg_price  carat_bin        cut
  0     863.908535          0      Ideal
  1    4213.864948          1      Ideal
  2   12838.984078          2      Ideal
  ...
  27  13466.823529          3       Fair
  28  15842.666667          4       Fair
  29  18018.000000          5       Fair
  """

  __name__ = "mutate"

  def __call__(self, df):
    for arg in self.args:
      if isinstance(arg, Later):
        df[str(arg)] = arg.evaluate(df, special="transform")
      else:
        df[str(arg)] = arg

    for key, val in _dict_to_possibly_ordered_tuples(self.kwargs):
      if isinstance(val, Later):
        df[key] = val.evaluate(df, special="transform")
      else:
        df[key] = val
    return df

  def __rrshift__(self, other):
    return self.__call__(DplyFrame(other.copy(deep=True)))


@ApplyToDataframe
def group_by(*args, **kwargs):
  def GroupDF(df):
    group_columns = list(kwargs.keys())
    mutate_columns = kwargs
    for arg in args:
      if arg._name is not None:
        group_columns.append(arg._name)
      else:
        group_columns.append(str(arg))
        mutate_columns[str(arg)] = arg

    if mutate_columns:
      df = df >> mutate(**mutate_columns)
    df.group_self(group_columns)
    return df
  return GroupDF


class summarize(Verb):
  """Summarizes a dataset via functions
  >>>(diamonds >>
  ...        group_by(X.cut, X.carat_bin) >>
  ...        summarize(avg_price=X.price.mean()))
  If the dataset has grouping, summarizing will be done for each group,
  otherwise, will return a single row
  """

  def __call__(self, df):
    def summarize(df):
      if df._grouped_on:
        input_dict = {k: val.evaluate(df, special="agg")
            for k, val in six.iteritems(self.kwargs)}
        return DplyFrame(input_dict).reset_index()
      else:
        input_dict = {k: val.evaluate(df) for k, val
                      in six.iteritems(self.kwargs)}

      if len(input_dict) == 0:
        return DplyFrame({}, index=index)
      if hasattr(df, "_current_group") and df._current_group:
        input_dict.update(df._current_group)
      index = [0]
      return DplyFrame(input_dict, index=index)

    outDf = summarize(df)

    # Remove multi-index created from grouping and applying
    for grouped_name in outDf.index.names[:-1]:
      if grouped_name in outDf:
        outDf.reset_index(level=0, drop=True, inplace=True)
      else:
        outDf.reset_index(level=0, inplace=True)

    # Drop all 0 index, created by summarize
    if (outDf.index == 0).all():
      outDf.reset_index(drop=True, inplace=True)
    return outDf >> ungroup()


@ApplyToDataframe
def count(*args, **kwargs):
  def CreateCountDf(df):
    col_name = df.columns[0]
    return (df >> group_by(*args, **kwargs) >>
                  summarize(n=X[col_name].__len__()) >> ungroup())
  return CreateCountDf
  

def UngroupDF(df):
  # df._grouped_on = None
  # df._group_dict = None
  df.ungroup()
  return df


@ApplyToDataframe
def ungroup():
  return UngroupDF


@ApplyToDataframe
def arrange(*args):
  """Sort DataFrame by the input column arguments.

  >>> diamonds >> sample_n(5) >> arrange(X.price) >> select(X.depth, X.price)
  Out:
         depth  price
  28547   61.0    675
  35132   59.1    889
  42526   61.3   1323
  3468    61.6   3392
  23829   62.0  11903
  """
  names = [column._name for column in args]
  def f(df):
    sortby_df = df >> mutate(*args)
    index = sortby_df.sort_values([str(arg) for arg in args]).index
    return df.loc[index]
  return f


@ApplyToDataframe
def head(*args, **kwargs):
  """Returns first n rows"""
  return lambda df: df.head(*args, **kwargs)


@ApplyToDataframe
def tail(*args, **kwargs):
  """Returns last n rows"""
  return lambda df: df.tail(*args, **kwargs)


@ApplyToDataframe
def sample_n(n):
  """Randomly sample n rows from the DataFrame"""
  return lambda df: DplyFrame(df.sample(n))


@ApplyToDataframe
def sample_frac(frac):
  """Randomly sample `frac` fraction of the DataFrame"""
  return lambda df: DplyFrame(df.sample(frac=frac))


@ApplyToDataframe
def sample(*args, **kwargs):
  """Convenience method that calls into pandas DataFrame's sample method"""
  return lambda df: df.sample(*args, **kwargs)


@ApplyToDataframe
def nrow():
  return lambda df: len(df)


@ApplyToDataframe
def rename(**kwargs):
  """Rename one or more columns, leaving other columns unchanged

  Example usage:
    diamonds >> rename(new_name=old_name)
  """
  def rename_columns(df):
    column_assignments = {old_name_later._name: new_name
                          for new_name, old_name_later in kwargs.items()}
    return df.rename(columns=column_assignments)
  return rename_columns


@ApplyToDataframe
def transmute(*args, **kwargs):
  """ Similar to `select` but allows mutation in column definitions.

  In : (diamonds >>
          head(3) >>
          transmute(new_price=X.price * 2, x_plus_y=X.x + X.y))
  Out:
        new_price  x_plus_y
    0        652      7.93
    1        652      7.73
    2        654      8.12
  """
  mutate_dateframe_fn = mutate(*args, **dict(kwargs))
  column_names_args = [str(arg) for arg in args]
  column_names_kwargs = [name for name, _
                         in _dict_to_possibly_ordered_tuples(kwargs)]
  column_names = column_names_args + column_names_kwargs
  return lambda df: mutate_dateframe_fn(df)[column_names]


@DelayFunction
def PairwiseGreater(series1, series2):
  index = series1.index
  newSeries = pandas.Series([max(s1, s2) for s1, s2 in zip(series1, series2)])
  newSeries.index = index
  return newSeries


@DelayFunction
def if_else(bool_series, series_true, series_false):
  index = bool_series.index
  newSeries = pandas.Series([s1 if b else s2 for b, s1, s2
      in zip(bool_series, series_true, series_false)])
  newSeries.index = index
  return newSeries


def get_join_cols(by_entry):
  """ helper function used for joins
  builds left and right join list for join function
  """
  left_cols = []
  right_cols = []
  for col in by_entry:
    if isinstance(col, str):
      left_cols.append(col)
      right_cols.append(col)
    else:
      left_cols.append(col[0])
      right_cols.append(col[1])
  return left_cols, right_cols


def mutating_join(*args, **kwargs):
  """ generic function for mutating dplyr-style joins
  """
  # candidate for improvement
  left = args[0]
  right = args[1]
  if 'by' in kwargs:
    left_cols, right_cols = get_join_cols(kwargs['by'])
  else:
    left_cols, right_cols = None, None
  if 'suffixes' in kwargs:
    dsuffixes = kwargs['suffixes']
  else:
    dsuffixes = ('_x', '_y')
  if left._grouped_on:
    outDf = (DplyFrame((left >> ungroup())
                       .merge(right, how=kwargs['how'], left_on=left_cols, 
                              right_on=right_cols, suffixes=dsuffixes))
             .regroup(left._grouped_on))
  else:
    outDf = DplyFrame(left.merge(right, how=kwargs['how'], left_on=left_cols, 
                                 right_on=right_cols, suffixes=dsuffixes))
  return outDf


class Join(Verb):
  """ Generic class for two-table verbs
  """

  def __new__(cls, *args, **kwargs):
    if (len(args) > 1 and 
        isinstance(args[0], pandas.DataFrame) and 
        isinstance(args[1], pandas.DataFrame)):
      verb = cls(*args[1:], **kwargs)
      return verb(args[0].copy(deep=True))
    else:
      return super(Verb, cls).__new__(cls)

  def __rrshift__(self, other):
      return self.__call__(other)


class inner_join(Join):
  """ Perform sql style inner join
  >>> left_data >> inner_join(right_data[
  ...                            , by=[join_columns_in_list_as_single_or_tuple][
  ...                            , suffixes=('_x', _y)]])
  e.g. flights2 >> inner_join(airports, by=[('origin', 'faa')]) >> head(5)

  returns dataframe preserving any grouping from left dataframe

  Select all rows from both tables where there are matches on specified columns.

  The by argument takes a list of columns. For a list like ['A', 'B'], it 
  assumes 'A' and 'B' are columns in both dataframes.

  For a list like [('A', 'B')], it assumes column 'A' in the left dataframe is 
  the same as column 'B' in the right dataframe.

  Can mix and match (e.g. by=['A', ('B', 'C')] will assume both dataframes have 
  column 'A', and column 'B' in the left dataframe is the same as column 'C' 
  in the right dataframe.

  If by is not specified, then all shared columns will be assumed to be the 
  join columns.

  suffixes will be used to rename columns that are common to both dataframes,
  but not used in the join operation.
  e.g. `suffixes=('_1', '_2')`

  If suffixes is not included, then the pandas default will be used ('_x', '_y')
  """
  __name__ = 'inner_join'

  def __call__(self, df):
    self.kwargs.update({'how': 'inner'})
    return mutating_join(df, self.args[0], **self.kwargs)


class full_join(Join):
  """ Perform sql style outer/full join
  >>> left_data >> full_join(right_data[
  ...                            , by=[join_columns_in_list_as_single_or_tuple][
  ...                            , suffixes=('_x', _y)]])
  e.g. flights2 >> full_join(airports, by=[('origin', 'faa')]) >> head(5)

  returns dataframe preserving any grouping from left dataframe

  Select all rows from both tables, matching when possible, filling in 
  missing values where data doesn't match.

  The by argument takes a list of columns. For a list like ['A', 'B'], it 
  assumes 'A' and 'B' are columns in both dataframes.
  
  For a list like [('A', 'B')], it assumes column 'A' in the left dataframe 
  is the same as column 'B' in the right dataframe.

  Can mix and match (e.g. by=['A', ('B', 'C')] will assume both dataframes 
  have  column 'A', and column 'B' in the left dataframe is the same as column 
  'C' in the right dataframe. If by is not specified, then all shared columns 
  will be assumed to be the join columns.

  suffixes will be used to rename columns that are common to both dataframes,
  but not used in the join operation.
  e.g. `suffixes=('_1', '_2')`.
  If suffixes is not included, then the pandas default will be used ('_x', '_y')
  """

  __name__ = 'full_join'

  def __call__(self, df):
    self.kwargs.update({'how': 'outer'})
    return mutating_join(df, self.args[0], **self.kwargs)


class left_join(Join):
  """ Perform sql style left join
  >>> left_data >> left_join(right_data[
  ...                            , by=[join_columns_in_list_as_single_or_tuple][
  ...                            , suffixes=('_x', _y)]])
  e.g. flights2 >> full_join(airports, by=[('origin', 'faa')]) >> head(5)

  returns dataframe preserving any grouping from left dataframe

  Select all rows from the left table, and corresponding rows from the right 
  table where values match, filling in missing values where data doesn't match.
  

  The by argument takes a list of columns. For a list like ['A', 'B'], it 
  assumes 'A' and 'B' are columns in both dataframes. For a list like [('A', 
  'B')], it assumes column 'A' in the left dataframe is the same as column 'B' 
  in the right dataframe. Can mix and match (e.g. by=['A', ('B', 'C')] will 
  assume both dataframes have column 'A', and column 'B' in the left dataframe 
  is the same as column 'C' in the right dataframe. If by is not specified, then
  all shared columns will be assumed to be the join columns.

  suffixes will be used to rename columns that are common to both dataframes,
  but not used in the join operation.
  e.g. `suffixes=('_1', '_2')`.
  If suffixes is not included, then the pandas default will be used ('_x', '_y')
  """

  __name__ = 'left_join'

  def __call__(self, df):
    self.kwargs.update({'how': 'left'})
    return mutating_join(df, self.args[0], **self.kwargs)


class right_join(Join):
  """ Perform sql style right join
  >>> left_data >> right_join(right_data[
  ...                            , by=[join_columns_in_list_as_single_or_tuple][
  ...                            , suffixes=('_x', _y)]])
  e.g. flights2 >> right_join(airports, by=[('origin', 'faa')]) >> head(5)

  returns dataframe preserving any grouping from left dataframe

  Select all rows from the right table, and corresponding rows from the left 
  table where values match, filling in missing values where data doesn't match.
  

  The by argument takes a list of columns. For a list like ['A', 'B'], it 
  assumes 'A' and 'B' are columns in both dataframes. For a list like [('A', 
  'B')], it assumes column 'A' in the left dataframe is the same as column 'B' 
  in the right dataframe. Can mix and match (e.g. by=['A', ('B', 'C')] will 
  assume both dataframes have column 'A', and column 'B' in the left dataframe 
  is the same as column 'C' in the right dataframe. If by is not specified, then
  all shared columns will be assumed to be the join columns.

  suffixes will 
  be used to rename columns that are common to both dataframes, but not used in 
  the join operation.
  e.g. `suffixes=('_1', '_2')`.
  If suffixes is not included, then the pandas default will be used ('_x', '_y')
  """

  __name__ = 'right_join'

  def __call__(self, df):
    self.kwargs.update({'how': 'right'})
    return mutating_join(df, self.args[0], **self.kwargs)


def filtering_join(*args, **kwargs):
  left = args[0]
  right = args[1]
  if 'by' in kwargs:
    left_cols, right_cols = get_join_cols(kwargs['by'])
    cols = lambda right, left: right_cols
  else:
    left_cols, right_cols = None, None
    cols = lambda right, left: [x for x in left.columns.values.tolist() if x in right.columns.values.tolist()]
  if left._grouped_on:
    outDf = DplyFrame((left >> ungroup())
                      .merge(right[cols(left, right)].drop_duplicates(), 
                             how=kwargs['how'], left_on=left_cols, 
                             right_on=right_cols, indicator=True, 
                             suffixes=('', '_y'))
                      .query(kwargs['query'])
                      .regroup(left._grouped_on)
                      .iloc[:, range(0, len(left.columns))])
  else:
    outDf = DplyFrame(left.merge(right[cols(left, right)]
                                 .drop_duplicates(), how=kwargs['how'], 
                                 left_on=left_cols, right_on=right_cols,
                                 indicator=True, suffixes=('', '_y'))
                      .query(kwargs['query'])
                      .iloc[:, range(0, len(left.columns))])
  return outDf


class semi_join(Join):
  """ Perform filtering semi join
  >>> left_data >> semi_join(right_data[
  ...                                    , by=[join_columns_in_list_as_single_or_tuple]])
  e.g. flights2 >> semi_join(airports, by=[('origin', 'faa')]) >> head(5)

  returns dataframe preserving any grouping from left dataframe

  Filters the left table by including observations that are also found in the
  right table. Never returns more observations than are found in the left table
  (i.e. multiple keys in the right table won't cause duplicate observations to 
  appear in the right table).
 
  The by argument takes a list of columns. For a
  list like ['A', 'B'], it assumes 'A' and 'B' are columns in both dataframes. 
  For a list like [('A', 'B')], it assumes column 'A' in the left dataframe is 
  the same as column 'B' in the right dataframe. Can mix and match (e.g. 
  by=['A', ('B', 'C')] will assume both dataframes have column 'A', and column 
  'B' in the left dataframe is the same as column 'C' in the right dataframe. If
  by is not specified, then all shared columns will be assumed to be the join 
  columns.
  
  """

  __name__ = 'semi_join'

  def __call__(self, df):
    self.kwargs.update({'how': 'inner', 'query': '_merge=="both"'})
    return filtering_join(df, self.args[0], **self.kwargs)


class anti_join(Join):
  """ Perform filtering anti join
  >>> left_data >> anti_join(right_data[, by=[join_columns_in_list_as_single_or_tuple]])
  e.g. flights2 >> anti_join(airports, by=[('origin', 'faa')]) >> head(5)

  returns dataframe with any grouping preserved from left dataframe

  Filters the left table by including observations that are not found in the 
  right table. Never returns more observations than are found in the left table 
  (i.e. multiple keys in the right table won't cause duplicate observations to 
  appear in the right table).

  The by argument takes a list of columns. For a 
  list like ['A', 'B'], it assumes 'A' and 'B' are columns in both dataframes. 
  For a list like [('A', 'B')], it assumes column 'A' in the left dataframe is 
  the same as column 'B' in the right dataframe. Can mix and match (e.g. 
  by=['A', ('B', 'C')] will assume both dataframes have column 'A', and column 
  'B' in the left dataframe is the same as column 'C' in the right dataframe. If
   by is not specified, then all shared columns will be assumed to be the join 
  columns. 
  """

  __name__ = 'anti_join'

  def __call__(self, df):
    self.kwargs.update({'how': 'left', 'query': '_merge=="left_only"'})
    return filtering_join(df, self.args[0], **self.kwargs)
